import fs from "fs";
import { deploy } from "./deploy";
import { cv } from "./cv";
import { faceidTestData } from "./test_data/faceid";
import path from "path";
import crypto from "crypto";

async function cacheImage(group: string, url: string) {
	const _url = new URL(url);
	const cacheDir = path.join(__dirname, "../cache/images", group);
	fs.mkdirSync(cacheDir, { recursive: true });
	const cacheJsonFile = path.join(cacheDir, "config.json");
	let jsonData: Record<string, string> = {};
	if (cacheJsonFile && fs.existsSync(cacheJsonFile)) {
		jsonData = JSON.parse(fs.readFileSync(cacheJsonFile, "utf-8"));
		const filename = jsonData[url];
		if (filename && fs.existsSync(filename)) return path.join(cacheDir, filename);
	}
	const data = await fetch(_url).then(res => res.arrayBuffer()).then(buf => new Uint8Array(buf));
	const allowedExtnames = [".jpg", ".jpeg", ".png", ".webp"];
	let extname = path.extname(_url.pathname);
	if (!allowedExtnames.includes(extname)) extname = ".jpeg";
	const md5 = crypto.hash("md5", data, "hex");
	const savename = md5 + extname;
	jsonData[url] = savename;
	fs.writeFileSync(cacheJsonFile, JSON.stringify(jsonData));
	const savepath = path.join(cacheDir, savename);
	fs.writeFileSync(savepath, data);
	return savepath;
}

async function testGenderTest() {
	const facedet = deploy.facedet.Yolov5Face.fromOnnx(fs.readFileSync("models/facedet/yolov5s.onnx"));
	const detector = deploy.faceattr.GenderAgeDetector.fromOnnx(fs.readFileSync("models/faceattr/insight_gender_age.onnx"));

	const image = await cv.Mat.load("https://b0.bdstatic.com/ugc/iHBWUj0XqytakT1ogBfBJwc7c305331d2cf904b9fb3d8dd3ed84f5.jpg");
	const boxes = await facedet.predict(image);
	if (!boxes.length) return console.error("未检测到人脸");
	for (const [idx, box] of boxes.entries()) {
		const res = await detector.predict(image, { crop: { sx: box.left, sy: box.top, sw: box.width, sh: box.height } });
		console.log(`[${idx + 1}]`, res);
	}

}

async function testFaceID() {
	const facedet = deploy.facedet.Yolov5Face.fromOnnx(fs.readFileSync("models/facedet/yolov5s.onnx"));
	const faceid = deploy.faceid.CosFace.fromOnnx(fs.readFileSync("models/faceid/insightface/glint360k_cosface_r100.onnx"));

	const { basic, tests } = faceidTestData.stars;

	console.log("正在加载图片资源");
	const basicImage = await cv.Mat.load(await cacheImage("faceid", basic.image));
	const testsImages: Record<string, cv.Mat[]> = {};
	for (const [name, imgs] of Object.entries(tests)) {
		testsImages[name] = await Promise.all(imgs.map(img => cacheImage("faceid", img).then(img => cv.Mat.load(img))));
	}

	console.log("正在检测基本数据");
	const basicDetectedFaces = await facedet.predict(basicImage);
	const basicFaceIndex: Record<string, number> = {};
	for (const [name, [x, y]] of Object.entries(basic.faces)) {
		basicFaceIndex[name] = basicDetectedFaces.findIndex(box => box.x1 < x && box.x2 > x && box.y1 < y && box.y2 > y);
	}

	const basicEmbds: number[][] = [];
	for (const box of basicDetectedFaces) {
		const embd = await faceid.predict(basicImage, { crop: { sx: box.left, sy: box.top, sw: box.width, sh: box.height } });
		basicEmbds.push(embd);
	}

	console.log("正在进行人脸对比");
	for (const [name, image] of Object.entries(testsImages)) {
		console.log(`[${name}] 正在检测`);
		const index = basicFaceIndex[name];
		if (index < 0) {
			console.error(`[${name}] 不存在`);
			continue;
		}
		const basicEmbd = basicEmbds[index]

		for (const [idx, img] of image.entries()) {
			const box = await facedet.predict(img).then(boxes => boxes[0]);
			if (!box) {
				console.error(`[${idx + 1}] 未检测到人脸`);
				continue
			}

			const embd = await faceid.predict(img, { crop: { sx: box.left, sy: box.top, sw: box.width, sh: box.height } });

			const compareEmbds = basicEmbds.map(e => deploy.faceid.cosineDistance(e, embd));
			const max = Math.max(...compareEmbds);
			const min = Math.min(...compareEmbds);

			const distance = deploy.faceid.cosineDistance(basicEmbd, embd);
			console.log(`[${idx + 1}] [${(distance.toFixed(4) == max.toFixed(4)) ? '\x1b[102m成功' : '\x1b[101m失败'}\x1b[0m] 相似度:${distance.toFixed(4)}, 最大:${max.toFixed(4)}, 最小:${min.toFixed(4)}`);
		}
	}

}

async function testFaceAlign() {
	const fd = deploy.facedet.Yolov5Face.fromOnnx(fs.readFileSync("models/facedet/yolov5s.onnx"));
	// const fa = deploy.facealign.PFLD.fromOnnx(fs.readFileSync("models/facealign/pfld-106-lite.onnx"));
	const fa = deploy.facealign.FaceLandmark1000.fromOnnx(fs.readFileSync("models/facealign/FaceLandmark1000.onnx"));
	let image = await cv.Mat.load("https://bkimg.cdn.bcebos.com/pic/d52a2834349b033b5bb5f183119c21d3d539b6001712");
	image = image.rotate(image.width / 2, image.height / 2, 0);

	const face = await fd.predict(image).then(res => res[0].toSquare());
	const points = await fa.predict(image, { crop: { sx: face.left, sy: face.top, sw: face.width, sh: face.height } });

	points.points.forEach((point, idx) => {
		image.circle(face.left + point.x, face.top + point.y, 2);
	})
	// const point = points.getPointsOf("rightEye")[1];
	// image.circle(face.left + point.x, face.top + point.y, 2);
	fs.writeFileSync("testdata/xx.jpg", image.encode(".jpg"));

	let faceImage = image.rotate(face.centerX, face.centerY, -points.direction * 180 / Math.PI);
	faceImage = faceImage.crop(face.left, face.top, face.width, face.height);
	fs.writeFileSync("testdata/face.jpg", faceImage.encode(".jpg"));

	console.log(points);
	console.log(points.direction * 180 / Math.PI);
	debugger
}


async function test() {
	// testGenderTest();
	// testFaceID();
	testFaceAlign();
}

test().catch(err => {
	console.error(err);
	debugger
});